#pragma once

enum zap_type
{
    ZAP_THROW_FLAME,
    ZAP_THROW_FROST,
    ZAP_SLOW,
    ZAP_HASTE,
    ZAP_MAGIC_DART,
    ZAP_PARALYSE,
    ZAP_BOLT_OF_FIRE,
    ZAP_BOLT_OF_COLD,
    ZAP_CONFUSE,
    ZAP_INVISIBILITY,
    ZAP_DIG,
    ZAP_FIREBALL,
    ZAP_TELEPORT_OTHER,
    ZAP_LIGHTNING_BOLT,
    ZAP_POLYMORPH,
    ZAP_VENOM_BOLT,
    ZAP_BOLT_OF_DRAINING,
    ZAP_LEHUDIBS_CRYSTAL_SPEAR,
    ZAP_CHARMING,
    ZAP_PAIN,
    ZAP_STICKY_FLAME,
    ZAP_PYRE_ARROW,
    ZAP_DISPEL_UNDEAD,
    ZAP_BANISHMENT,
    ZAP_STING,
    ZAP_HURL_DAMNATION,
    ZAP_IRON_SHOT,
    ZAP_STONE_ARROW,
    ZAP_SHOCK,
    ZAP_ORB_OF_ELECTRICITY,
    ZAP_SPIT_POISON,
    ZAP_DEBUGGING_RAY,
    ZAP_BREATHE_FIRE,
    ZAP_GLACIAL_BREATH,
    ZAP_BREATHE_ACID,
    ZAP_BREATHE_POISON,
    ZAP_AGONY,
    ZAP_MINDBURST,
    ZAP_BREATHE_STEAM,
    ZAP_THROW_ICICLE,
    ZAP_CORONA,
    ZAP_HIBERNATION,
    ZAP_SANDBLAST,
    ZAP_BOLT_OF_MAGMA,
    ZAP_POISON_ARROW,
    ZAP_PETRIFY,
    ZAP_PORKALATOR,
    ZAP_SLEEP,
    ZAP_PRIMAL_WAVE,
    ZAP_IOOD,
    ZAP_NOXIOUS_BREATH,
    ZAP_INNER_FLAME,
    ZAP_FORCE_LANCE,
    ZAP_SEARING_RAY,
    ZAP_REBOUNDING_BLAZE,
    ZAP_REBOUNDING_CHILL,
    ZAP_TUKIMAS_DANCE,
    ZAP_QUICKSILVER_BOLT,
    ZAP_CORROSIVE_BOLT,
    ZAP_MEPHITIC,
    ZAP_UNRAVELLING,
    ZAP_ICEBLAST,
    ZAP_SLUG_DART,
    ZAP_BLINKBOLT,
    ZAP_FREEZING_GUST,
    ZAP_DIMENSION_ANCHOR,
    ZAP_VULNERABILITY,
    ZAP_SENTINEL_MARK,
    ZAP_SAP_MAGIC,
    ZAP_VIRULENCE,
    ZAP_BECKONING,
    ZAP_FIRE_STORM,
    ZAP_MALMUTATE,
    ZAP_VILE_CLUTCH,
    ZAP_HARPOON_SHOT,
    ZAP_REFRIGERATE,
    ZAP_DRAIN_LIFE,
    ZAP_SONIC_WAVE,
    ZAP_THROW_PIE,
    ZAP_FOXFIRE,
    ZAP_DISPEL_UNDEAD_RANGE,
    ZAP_TREMORSTONE,
    ZAP_NOXIOUS_CLOUD,
    ZAP_POISONOUS_CLOUD,
    ZAP_BOLT_OF_DEVASTATION,
    ZAP_METAL_SPLINTERS,
    ZAP_SPLINTERSPRAY,
    ZAP_THORN_VOLLEY,
    ZAP_CALL_DOWN_DAMNATION,
    ZAP_MIASMA_BREATH,
    ZAP_PETRIFYING_CLOUD,
    ZAP_GHOSTLY_FIREBALL,
    ZAP_FLASH_FREEZE,
    ZAP_CREEPING_FROST,
    ZAP_SPIT_LAVA,
    ZAP_THROW_BARBS,
    ZAP_HURL_SLUDGE,
    ZAP_HOLY_BREATH,
    ZAP_SPIT_ACID,
    ZAP_CHAOS_BREATH,
    ZAP_COLD_BREATH,
    ZAP_FIRE_BREATH,
    ZAP_SEARING_BREATH,
    ZAP_ELECTRICAL_BOLT,
    ZAP_FLAMING_CLOUD,
    ZAP_IGNITION,
    ZAP_HAILSTORM,
    ZAP_MYSTIC_BLAST,
    ZAP_SWORD_BEAM,
    ZAP_MIGHT,
    ZAP_VAMPIRIC_DRAINING,
    ZAP_CHAIN_LIGHTNING,
    ZAP_STUNNING_BURST,
    ZAP_FLAME_WAVE,
    ZAP_ENFEEBLE,
    ZAP_SOUL_SPLINTER,
    ZAP_BOLT_OF_LIGHT,
    ZAP_FASTROOT,
    ZAP_BLASTMOTE,
    ZAP_KISS_OF_DEATH,
    ZAP_MOMENTUM_STRIKE,
    ZAP_PLASMA,
    ZAP_BOMBARD,
    ZAP_PLASMA_LIGHTNING,
    ZAP_VITRIFY,
    ZAP_MALIGN_OFFERING,
    ZAP_CURSE_OF_AGONY,
    ZAP_HURL_TORCHLIGHT,
    ZAP_CRYSTALLISING_SHOT,
    ZAP_WARP_SPACE,
    ZAP_COMBUSTION_BREATH,
    ZAP_NULLIFYING_BREATH,
    ZAP_MUD_BREATH,
    ZAP_GALVANIC_BREATH,
    ZAP_CAUSTIC_BREATH,
    ZAP_AUTODAZZLE,
    ZAP_PERMAFROST_ERUPTION_COLD,
    ZAP_PERMAFROST_ERUPTION_EARTH,
    ZAP_GRAVITAS,
    ZAP_MAGNAVOLT,
    ZAP_FULSOME_FUSILLADE,
    ZAP_RIMEBLIGHT,
    ZAP_RIMEBLIGHT_SHARDS,
    ZAP_HOARFROST_BULLET,
    ZAP_HOARFROST_BULLET_FINALE,
    ZAP_FLASHING_BALESTRA,
    ZAP_HELLFIRE_MORTAR_DIG,
    ZAP_SHADOW_SHARD,
    ZAP_SHADOW_BOLT,
    ZAP_SHADOWBALL,
    ZAP_CREEPING_SHADOW,
    ZAP_SHADOW_JOLT,
    ZAP_SHADOW_DRAIN,
    ZAP_SHADOW_TORPOR,
    ZAP_GRAVE_CLAW,
    ZAP_UNLEASH_DESTRUCTION,
    ZAP_UNLEASH_DESTRUCTION_PIERCING,
    ZAP_BLOOD_ARROW,
    ZAP_PHANTOM_BLITZ,
    ZAP_SPIKE_LAUNCHER,
    ZAP_KINETIC_GRAPNEL,
    ZAP_SHRED,
    ZAP_PARAGON_IMPACT,
    ZAP_MONARCH_DETONATION,
    ZAP_DEPLOY_BOMBLET,
    ZAP_SPLINTERFROST_FRAGMENT,
    ZAP_PERCUSSIVE_TEMPERING,
    ZAP_FORTRESS_BLAST,
    ZAP_SEISMIC_STOMP,
    ZAP_RENDING_SLASH,
    ZAP_MAGMA_BARRAGE,
    ZAP_SHADOW_SHOT,
    ZAP_BOLAS,
    ZAP_MERCURY_ARROW,
    ZAP_SOJOURNING,
    ZAP_FREEZE,
    ZAP_VEX,
    ZAP_RAVENOUS_SWARM,
    ZAP_WATERY_GRAVE,
    ZAP_RUST_BREATH,
    ZAP_GOLDEN_BREATH,
    ZAP_BECKONING_GALE,
    ZAP_GHOSTLY_SACRIFICE,
    ZAP_BOLT_OF_FLESH,
    ZAP_AWAKEN_FLESH,
    ZAP_ILL_OMEN,
    ZAP_DOOM_BOLT,
    ZAP_WARP_BODY,
    ZAP_CONTAM_EXPLOSION,
    ZAP_ACID_BALL,
    ZAP_SHOOTING_STAR,
    ZAP_DIMINISH_SPELLS,
    ZAP_VALOUR_BEAM,
    ZAP_LANDBREAKER,
    ZAP_BURSTSPORE,
    NUM_ZAPS
};
