del ..\rltiles\gui.d
del ..\rltiles\gui.png
del ..\rltiles\icons.d
del ..\rltiles\icons.png
del ..\rltiles\player.d
del ..\rltiles\player.png
del ..\rltiles\tiledef-gui.cc
del ..\rltiles\tiledef-gui.h
del ..\rltiles\tiledef-icons.cc
del ..\rltiles\tiledef-icons.h
del ..\rltiles\tiledef-player.cc
del ..\rltiles\tiledef-player.h
del ..\rltiles\tile-gui.html
del ..\rltiles\tile-icons.html
del ..\rltiles\tileinfo-gui.js
del ..\rltiles\tileinfo-icons.js
del ..\rltiles\tileinfo-player.js
del ..\rltiles\tile-player.html
del ..\rltiles\feat.d
del ..\rltiles\feat.png
del ..\rltiles\tiledef-feat.cc
del ..\rltiles\tiledef-feat.h
del ..\rltiles\tiledef-wall.cc
del ..\rltiles\tiledef-wall.h
del ..\rltiles\tile-feat.html
del ..\rltiles\tileinfo-feat.js
del ..\rltiles\tileinfo-wall.js
del ..\rltiles\tile-wall.html
del ..\rltiles\wall.d
del ..\rltiles\wall.png
del ..\rltiles\dngn.png
del ..\rltiles\floor.d
del ..\rltiles\floor.png
del ..\rltiles\main.d
del ..\rltiles\main.png
del ..\rltiles\tiledef-dngn.cc
del ..\rltiles\tiledef-dngn.h
del ..\rltiles\tiledef-floor.cc
del ..\rltiles\tiledef-floor.h
del ..\rltiles\tiledef-main.cc
del ..\rltiles\tiledef-main.h
del ..\rltiles\tile-floor.html
del ..\rltiles\tile-main.html
del ..\rltiles\tileinfo-dngn.js
del ..\rltiles\tileinfo-floor.js
del ..\rltiles\tileinfo-main.js
pause
