/** \class actor
    \brief Virtual base class for player and monster classes.

    Contains all the properties and methods common to monsters and players.
*/

#pragma once

/** \var CrawlHashTable actor::props
    \brief Properties hash table.

    actor keys
    public
    private

    player keys
    public
    private

    monster keys
    public
    private
*/
