#pragma once

#include "tag-version.h"

enum trap_type
{
#if TAG_MAJOR_VERSION == 34
    TRAP_HARLEQUIN,
    TRAP_FIRST_TRAP = TRAP_HARLEQUIN,
    TRAP_ARCHMAGE,
    TRAP_SPEAR,
#endif
#if TAG_MAJOR_VERSION > 34
    TRAP_DISPERSAL,
    TRAP_FIRST_TRAP = TRAP_DISPERSAL,
    TRAP_TELEPORT,
#endif
    TRAP_TELEPORT_PERMANENT,
#if TAG_MAJOR_VERSION > 34
    TRAP_TYRANT,
    TRAP_ARCHMAGE,
    TRAP_HARLEQUIN,
    TRAP_DEVOURER,
#endif
    TRAP_ALARM,
#if TAG_MAJOR_VERSION == 34
    TRAP_TYRANT,
    TRAP_BOLT,
#endif
    TRAP_NET,
    TRAP_ZOT,
#if TAG_MAJOR_VERSION == 34
    TRAP_DEVOURER,
#endif
    TRAP_SHAFT,
    TRAP_GOLUBRIA,
    TRAP_PLATE,
    TRAP_WEB,
#if TAG_MAJOR_VERSION == 34
    TRAP_GAS,
    TRAP_TELEPORT,
    TRAP_SHADOW,
    TRAP_SHADOW_DORMANT,
    TRAP_DISPERSAL,
#endif
    NUM_TRAPS,
    TRAP_MAX_REGULAR = TRAP_SHAFT,
    TRAP_UNASSIGNED = 100,
#if TAG_MAJOR_VERSION == 34
    TRAP_UNUSED1,                      // was TRAP_INDEPTH
    TRAP_UNUSED2,                      // was TRAP_NOTELEPORT
#endif
    TRAP_RANDOM,
};
