
#pragma once

enum deck_type
{
    DECK_OF_ESCAPE,
    FIRST_PLAYER_DECK = DECK_OF_ESCAPE,
    DECK_OF_DESTRUCTION,
    DECK_OF_SUMMONING,
    LAST_PLAYER_DECK = DECK_OF_SUMMONING,
    DECK_OF_PUNISHMENT,
    DECK_STACK,
    NUM_DECKS
};
