#pragma once

enum lookup_help_type
{
    LOOKUP_HELP_MONSTER,
    FIRST_LOOKUP_HELP_TYPE = LOOKUP_HELP_MONSTER,
    LOOKUP_HELP_SPELL,
    LOOKUP_HELP_SKILL,
    LOOKUP_HELP_ABILITY,
    LOOKUP_HELP_CARD,
    LOOKUP_HELP_ITEM,
    LOOKUP_HELP_FEATURE,
    LOOKUP_HELP_GOD,
    LOOKUP_HELP_BRANCH,
    LOOKUP_HELP_CLOUD,
    LOOKUP_HELP_PASSIVE,
    LOOKUP_HELP_STATUS,
    LOOKUP_HELP_MUTATION,
    LOOKUP_HELP_BANE,
    NUM_LOOKUP_HELP_TYPES
};
