#pragma once

enum terrain_change_type
{
    TERRAIN_CHANGE_GENERIC,
    TERRAIN_CHANGE_FLOOD,
    TERRAIN_CHANGE_TOMB,
    TERRAIN_CHANGE_IMPRISON,
    TERRAIN_CHANGE_DOOR_SEAL,
    TERRAIN_CHANGE_FORESTED,
    TERRAIN_CHANGE_BOG,
    TERRAIN_CHANGE_SLIME,
    TERRAIN_CHANGE_BINDING_SIGIL,
    TERRAIN_CHANGE_HELLFIRE_MORTAR,
    TERRAIN_CHANGE_WEBS,
    TERRAIN_CHANGE_SPIKE_LAUNCHER,
    TERRAIN_CHANGE_RIME_YAK,
    TERRAIN_CHANGE_AQUA_FORM,
    NUM_TERRAIN_CHANGE_TYPES
};
