#pragma once

// Cleansing flame sources
enum class cleansing_flame_source
{
    generic    = -1,
    spell      = -2, // SPELL_FLAME_OF_CLEANSING
    invocation = -3, // ABIL_TSO_CLEANSING_FLAME
    tso        = -4, // TSO effect
};
