#pragma once

enum curse_type
{
    CURSE_MELEE,
    CURSE_FIRST_CURSE = CURSE_MELEE,
    CURSE_RANGED,
    CURSE_ELEMENTS,
    CURSE_SORCERY,
    CURSE_COMPANIONS,
    CURSE_BEGUILING,
    CURSE_SELF,
    CURSE_FORTITUDE,
    CURSE_CUNNING,
    CURSE_DEVICES,
    NUM_CURSES,
};
